import struct
from dataclasses import Field, asdict, dataclass, field
from enum import IntEnum
from struct import unpack_from
from typing import Optional, Union


class ButtonId(IntEnum):
    """Which button is a package representing"""

    TOP = 0
    MIDDLE = 1
    BOTTOM = 2


class ButtonAction(IntEnum):
    """The action of a button package

    Up: button was released
    Down: button was pressed down
    Long: a long press was detected (button down for 500ms or more)
    LongUp: a long press was released (I think you'll always receive an "up" event with this one)
    ExtraLong: an extra long press was detected (I think it's about 2000ms, you'll always receive a "long" event
               before this one)
    ExtraLongUp: button released after an "extra long" press
    Click: button was pressed down and released "quickly" (i.e. not triggering a long press)
    DoubleClick: not used
    """

    UP = 0
    DOWN = 1
    LONG = 2
    LONGUP = 3
    EXTRALONG = 4
    EXTRALONGUP = 5
    CLICK = 6
    DOUBLECLICK = 7


@dataclass(frozen=True)
class Point3d:
    x: float
    y: float
    z: float

    def __sub__(self, other):
        return Point3d(x=self.x - other.x, y=self.y - other.y, z=self.z - other.z)


@dataclass(frozen=True)
class Point4d:
    w: float
    x: float
    y: float
    z: float


@dataclass(frozen=True)
class Euler3d:
    roll: float
    pitch: float
    yaw: float


class DeviceMode(IntEnum):
    PRESET = 100
    SOFTWAVE = 101
    WAVEFRONT = 102
    API = 103


class PackageType(IntEnum):
    REQUEST = 1
    RESPONSE = 2
    STREAM = 3


class PackageId(IntEnum):
    DATASTREAM = 1
    BATTERY_STATUS = 2
    DEVICE_INFO = 3
    BUTTON_EVENT = 4
    DEVICE_MODE = 5
    IDENTIFY = 6
    RECENTER = 7
    DISPLAY_FRAME = 8


@dataclass(frozen=True)
class PackageMetadata:
    """
    This is called `Query` in the public types header file. Splits out and holds the metadata from each package received
    """

    _fmt = "<BBH"

    type: PackageType
    id: PackageId
    payload_size: int

    @classmethod
    def from_raw_bytes(cls, raw_bytes: Union[bytearray, bytes]) -> "PackageMetadata":
        # TODO(robert): Check annotation `bytes`
        q_type, q_id, q_payload_size = unpack_from(cls._fmt, raw_bytes, 0)
        return cls(type=q_type, id=q_id, payload_size=q_payload_size)

    @classmethod
    def split_out_data_from_metadata(cls, raw_bytes: Union[bytearray, bytes]) -> Union[bytearray, bytes]:
        return raw_bytes[struct.calcsize(cls._fmt) :]

    def to_bytes(self) -> bytes:
        return struct.pack(self._fmt, self.type, self.id, self.payload_size)

    def is_data(self):
        return self.id == PackageId.DATASTREAM

    def is_button(self):
        return self.id == PackageId.BUTTON_EVENT


@dataclass(frozen=True)
class DataPackage:
    """Represents a data package sent from wave

    Note: Initializing DataClasses is (relatively) slow, if in the unlikely event this becomes a bottleneck,
          refactor into a new structure
    """

    _raw_len = 105

    gyro: Point3d
    acc: Point3d
    mag: Point3d
    raw_pose: Point4d
    current_pose: Point4d
    euler: Euler3d
    linacc: Point3d
    peak: bool
    peak_norm_velocity: float
    timestamp_us: int
    grav: Point3d = field(init=False)

    def __post_init__(self):
        # A way to initialize a derived field in a frozen dataclass
        super().__setattr__("grav", self.acc - self.linacc)

    @classmethod
    def from_raw_bytes(cls, data: Union[bytearray, bytes]) -> "DataPackage":
        # Explanation for the byte structure: https://docs.python.org/3/library/struct.html
        assert len(data) == cls._raw_len, f"Expected the raw data to have len={cls._raw_len}, got len={len(data)}"
        # These parameters encode how to read the bytes from the stream
        return cls(
            gyro=Point3d(*unpack_from("<3f", data, 0)),
            acc=Point3d(*unpack_from("<3f", data, 12)),
            mag=Point3d(*unpack_from("<3f", data, 24)),
            raw_pose=Point4d(*unpack_from("<4f", data, 36)),
            current_pose=Point4d(*unpack_from("<4f", data, 52)),
            euler=Euler3d(*unpack_from("<3f", data, 68)),
            linacc=Point3d(*unpack_from("<3f", data, 80)),
            peak=unpack_from("?", data, 92)[0],
            peak_norm_velocity=unpack_from("<f", data, 93)[0],
            timestamp_us=unpack_from("<Q", data, 97)[0],
        )

    def as_flat_dict(self) -> dict:
        # Recursively unpack into dicts
        d = asdict(self)
        d = flatten_nested_dicts(d, None)
        return d

    @classmethod
    def flat_keys(cls) -> tuple:
        return tuple(flatten_nested_dataclass_fields(cls, None))


@dataclass(frozen=True)
class ButtonEvent:
    """Represents a button event sent from wave"""

    _raw_len = 8
    _relevant_idx = 2

    button_id: ButtonId
    action: ButtonAction

    @classmethod
    def from_raw_bytes(cls, data: Union[bytearray, bytes]) -> "ButtonEvent":
        """Decomposes raw bytes into its components and returns them as a :class:`ButtonEvent`"""
        assert len(data) == cls._raw_len, f"Expected to get {cls._raw_len} bytes, got {len(data)}"
        button_id, action = unpack_from("<BB", data, 0)
        return cls(button_id=ButtonId(button_id), action=ButtonAction(action))


def flatten_nested_dicts(d: dict, name: Optional[str]) -> dict:
    """Recursively flattens dicts of dicts into a dict with keys concatenated

    Args:
        d: The dictionary to flatten
        name: The prefix to this level, or more specifically, the key for the next higher level. If `None`, no prefix
              is used

    Returns:
        A flat dict where the keys have been concatenated

    Examples:
        >>> flatten_nested_dicts({"a": {"b": {"c": 1}, "d": 2}, "e": 3}, None)
        {'a_b_c': 1, 'a_d': 2, 'e': 3}
    """
    d_results = {}
    if isinstance(d, dict):
        for k, v in d.items():
            curr_name = f"{name}_{k}" if name is not None else k
            curr_val = flatten_nested_dicts(v, curr_name)
            d_results.update(curr_val)
    else:
        curr_val = {name: d}
        d_results.update(curr_val)

    return d_results


def flatten_nested_dataclass_fields(d: Union[Field, type], name: Optional[str]) -> list:
    """Analogous to `flatten_nested_dicts`, but returns the key names and works on the static class, not an instance

    Args:
        d: A dataclass e.g. `DataPackage`
        name: The prefix to this level, or more specifically, the key for the next higher level. If `None`, no prefix
              is used

    Returns:
        Returns the concatenated keys

    Examples:
        >>> @dataclass
        ... class A:
        ...     x: float
        ...     y: float
        >>> flatten_nested_dataclass_fields(A, None)
        ['x', 'y']
        >>> @dataclass
        ... class B:
        ...     a: A
        ...     b: int
        >>> flatten_nested_dataclass_fields(B, None)
        ['a_x', 'a_y', 'b']
    """
    results = []
    if hasattr(d, "__dataclass_fields__"):
        for k, v in d.__dataclass_fields__.items():
            curr_name = f"{name}_{k}" if name is not None else k
            curr_val = flatten_nested_dataclass_fields(v, curr_name)
            results.extend(curr_val)
    elif isinstance(d, Field):
        curr_val = flatten_nested_dataclass_fields(d.type, name)
        results.extend(curr_val)
    else:
        results.append(name)

    return results


def process_byte_data(raw_bytes: Union[bytearray, bytes]) -> Union[ButtonEvent, DataPackage]:
    """Factory function that takes raw bytes and translates into a button event or data

    Args:
        raw_bytes: The input raw bytes

    Returns:
        The resulting button event or data
    """
    q = PackageMetadata.from_raw_bytes(raw_bytes)
    raw_bytes_data = PackageMetadata.split_out_data_from_metadata(raw_bytes)

    if q.is_data():
        package = DataPackage.from_raw_bytes(raw_bytes_data)
    elif q.is_button():
        package = ButtonEvent.from_raw_bytes(raw_bytes_data)
    else:
        raise ValueError(f"Unknown value for q.id={q.id}")

    return package



import asyncio
import logging
import signal
import sys
from functools import partial
from typing import Union, List, Callable, Tuple

import serial
from bleak import BleakClient
from serial_asyncio import open_serial_connection

from genki_wave.callbacks import WaveCallback
from genki_wave.constants import API_CHAR_UUID, BAUDRATE
from genki_wave.data.writing import get_start_api_package
from genki_wave.protocols import ProtocolAsyncio, ProtocolThread, CommunicateCancel
from genki_wave.utils import get_serial_port, get_or_create_event_loop

logging.basicConfig(format="%(levelname).4s:%(asctime)s [%(filename)s:%(lineno)d] - %(message)s ")
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def prepare_protocol_as_bleak_callback_asyncio(protocol: ProtocolAsyncio) -> Callable:
    async def _inner(sender: str, data: bytearray) -> None:
        # NOTE: `bleak` expects a function with this signature
        await protocol.data_received(data)

    return _inner


def prepare_protocol_as_bleak_callback(protocol: ProtocolThread) -> Callable:
    def _inner(sender: str, data: bytearray) -> None:
        # NOTE: `bleak` expects a function with this signature
        protocol.data_received(data)

    return _inner


def bleak_callback(protocol: ProtocolAsyncio) -> Callable:
    """Wraps our protocol as a callback with the correct signature bleak expects

    NOTE: 1) Bleak checks if a function is a co-routine so we need to wrap the class method into an `async` function
          and 2) we need to take care that `asyncio.Queue` is correctly handled so we have 2 different wrappers, one
          for a regular `queue.Queue` and one for `asyncio.Queue`.
    """
    if isinstance(protocol, ProtocolAsyncio):
        callback = prepare_protocol_as_bleak_callback_asyncio(protocol)
    elif isinstance(protocol, ProtocolThread):
        callback = prepare_protocol_as_bleak_callback(protocol)
    else:
        raise ValueError(f"Unknown protocol type {type(protocol)}")
    return callback


def make_disconnect_callback(comm: CommunicateCancel):
    def cb(client):
        if not comm.cancel:
            print(f"Client {client.address} disconnected unexpectedly, exiting")
            sys.exit(1)

    return cb


async def producer_bluetooth(
    protocol: Union[ProtocolAsyncio, ProtocolThread],
    comm: CommunicateCancel,
    ble_address: str,
) -> None:
    """Receives data from a serially connected wave ring and passes it to the `protocol`

    Args:
        protocol: An object that knows how to process the raw data sent from the Wave ring into a structured format
                  and passes it along between `producer` and `consumer`.
        comm: An object that allows `producer` and `consumer` to communicate when to cancel the process
        ble_address: Address of the bluetooth device to connect to. E.g. 'D5:73:DB:85:B4:A1'

    Note:
        The producer doesn't return a value, but the data gets added to the `protocol` that can be accessed from other
        parts of the program i.e. some `consumer`
    """
    print(f"Connecting to wave at address {ble_address}")
    callback = bleak_callback(protocol)
    async with BleakClient(ble_address, disconnected_callback=make_disconnect_callback(comm)) as client:
        await client.start_notify(API_CHAR_UUID, callback)
        await client.write_gatt_char(API_CHAR_UUID, get_start_api_package(), False)

        print("Connected to Wave")
        while True:
            # This `while` loop and `asyncio.sleep` statement is some magic that is required to continually fetch
            # the data from the bluetooth device.
            await asyncio.sleep(0.1)

            if comm.cancel:
                print("Recieved a cancel signal, stopping ble client")
                break

        await client.stop_notify(API_CHAR_UUID)


async def producer_serial(protocol: ProtocolAsyncio, comm: CommunicateCancel, serial_port: str):
    """Receives data from a serially connected wave ring and passes it to the `protocol`

    Args:
        protocol: An object that knows how to process the raw data sent from the Wave ring into a structured format
                  and passes it along between `producer` and `consumer`.
        comm: An object that allows `producer` and `consumer` to communicate when to cancel the process
        serial_port: The serial port to read from

    Note:
        The producer doesn't return a value, but the data gets added to the `protocol` that can be accessed from other
        parts of the program i.e. some `consumer`
    """
    reader, writer = await open_serial_connection(url=serial_port, baudrate=BAUDRATE, parity=serial.PARITY_EVEN)
    writer.write(get_start_api_package())
    while True:
        # The number of bytes read here is an arbitrary power of 2 on the order of a size of a single package
        packet = await reader.read(n=128)
        await protocol.data_received(packet)

        if comm.cancel:
            print("Recieved a cancel signal, stopping serial connection")
            break


async def consumer(
    protocol: ProtocolAsyncio,
    comm: CommunicateCancel,
    callbacks: Union[List[WaveCallback], Tuple[WaveCallback]],
) -> None:
    """Consumes the data from a producer via a protocol

    Args:
        protocol: An object that knows how to process the raw data sent from the Wave ring into a structured format
                  and passes it along between `producer` and `consumer`.
        comm: An object that allows `producer` and `consumer` to communicate when to cancel the process
        callbacks: A list/tuple of callbacks that handle the data passed from the wave ring when available
    """
    while True:
        package = await protocol.queue.get()

        if comm.is_cancel(package) or comm.cancel:
            print("Got a cancel message. Exiting consumer loop...")
            comm.cancel = True
            break

        for callback in callbacks:
            callback(package)


def make_sigint_handler(comm: CommunicateCancel):
    """Create a signal handler to cancel an asyncio loop using signals."""

    def handler(*args):
        comm.cancel = True

    return handler


def _run_asyncio(
    callbacks: List[WaveCallback], producer: Union[producer_bluetooth, producer_serial], protocol: ProtocolAsyncio
) -> None:
    """Runs a producer and a consumer, hooking into the data using the supplied callbacks

    Args:
        callbacks: See docs for `consumer`
        producer: A callable that takes 2 arguments, a protocol and a communication object
        protocol: An object that knows how to process the raw data sent from the Wave ring into a structured format
                  and passes it along between `producer` and `consumer`.
    """
    # A singleton that sends messages about whether the data transfer has been canceled.
    comm = CommunicateCancel()
    loop = get_or_create_event_loop()
    loop.add_signal_handler(signal.SIGINT, make_sigint_handler(comm))

    # Note: The consumer and the producer send the data via the instance of `protocol`
    tasks = asyncio.gather(producer(protocol, comm), consumer(protocol, comm, callbacks))
    loop.run_until_complete(tasks)


def run_asyncio_bluetooth(callbacks: List[WaveCallback], ble_address) -> None:
    """Runs an async `consumer-producer` loop using user supplied callbacks for a bluetooth device

    Args:
        callbacks: A list/tuple of callbacks that handle the data passed from the wave ring
        ble_address: Address of the bluetooth device to connect to. E.g. 'D5:73:DB:85:B4:A1'
    """
    _run_asyncio(callbacks, partial(producer_bluetooth, ble_address=ble_address), ProtocolAsyncio())


def run_asyncio_serial(callbacks: List[WaveCallback], serial_port: str = None) -> None:
    """Runs an async `consumer-producer` loop using user supplied callbacks for a serial device

    Args:
        callbacks: A list/tuple of callbacks that handle the data passed from the wave ring
        serial_port: The serial port to read from. If `None` will try to determine it automatically based on the
                     operating system the script is running on
    """
    serial_port = get_serial_port() if serial_port is None else serial_port

    _run_asyncio(callbacks, partial(producer_serial, serial_port=serial_port), ProtocolAsyncio())



import asyncio

import bleak


async def discover_bluetooth(device_str: str):
    print("Searching for bluetooth devices...")
    devices = await bleak.discover()
    wave_devices = [device for device in devices if device_str in device.name]
    print("Found the following devices" if wave_devices else "Found no devices")
    for device in wave_devices:
        print(f"Address: {device.address} - Name: {device.name}")


def run_discover_bluetooth():
    """Finds all `Wave` devices available for bluetooth connection and prints the addresses"""
    asyncio.get_event_loop().run_until_complete(discover_bluetooth("Wave"))



import argparse
from pathlib import Path

from genki_wave.callbacks import ButtonAndDataPrint, CsvOutput
from genki_wave.asyncio_runner import run_asyncio_bluetooth, run_asyncio_serial
from genki_wave.discover import run_discover_bluetooth


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--ble-address", type=str)
    parser.add_argument("--use-serial", action="store_true")
    parser.add_argument("--button", action="store_true", help="Handler that prints the button and the action")
    parser.add_argument(
        "--csv", type=str, default=None, help="Path to the output csv. If none is given no csv is written"
    )
    args = parser.parse_args()

    callbacks = []
    if args.button:
        callbacks.append(ButtonAndDataPrint(5))
    if args.csv is not None:
        callbacks.append(CsvOutput(Path(args.csv)))

    if not callbacks:
        print("Warning: no callbacks supplied, the data received won't be processed in any way")

    if args.use_serial:
        run_asyncio_serial(callbacks)
    else:
        if args.ble_address is None:
            print("No bluetooth address (--ble-address) supplied, searching for devices...")
            run_discover_bluetooth()
        else:
            print("Turn off by holding the `TOP` button")
            run_asyncio_bluetooth(callbacks, args.ble_address)


if __name__ == "__main__":
    main()



"""
This script is adopted from [here](https://github.com/atizo/pygame/blob/master/examples/midi.py)

# MacOS and Windows

Connecting the wave ring as a midi device via bluetooth should work out-of-the-box on Windows and MacOS.

# Linux
On Linux you might need to re-install `BlueZ` with support for midi as follows.
Installation instructions for Linux. Taken from
[here](https://tttapa.github.io/Pages/Ubuntu/Software-Installation/BlueZ.html)

```
sudo apt install libudev-dev libical-dev libreadline-dev libdbus-1-dev libasound2-dev
sudo apt install build-essential

# Use whatever version of BlueZ you want and change accordingly
cd /tmp
wget https://mirrors.edge.kernel.org/pub/linux/bluetooth/bluez-5.53.tar.xz
tar -xf bluez-5.53.tar.xz

cd bluez-5.53
./configure --enable-midi --with-systemdsystemunitdir=/etc/systemd/system
make
sudo make install
sudo apt-get install --reinstall bluez
```

You might need a restart after this step

Check if the device is detected:
```
aconnect -lio
```

If pygame raises a warning that `Cannot open shared library libasound_module_conf_pulse.so` it might be searching
in the wrong place a way to fix it is to create a symbolic link to where the actual library is to where pygame searches.
An example for ubuntu (pygame searches in `/usr/lib/alsa-lib`)
sudo ln -s /usr/lib/x86_64-linux-gnu/alsa-lib /usr/lib/alsa-lib

Solution adapted from [here](https://stackoverflow.com/questions/57946421/im-trying-to-run-this-code-initialize-pygame-midi-code-but-it-returns-an-error)  # noqa
"""
import functools
from enum import IntEnum
from typing import Optional, Tuple

try:
    import pygame
    import pygame.midi
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "This script only depends on pygame, not the genki-wave library so please install 'pygame'"
        "manually to run this script"
    )


class MidiStatus(IntEnum):
    DATA = 176


class Movements(IntEnum):
    PITCH = 16
    YAW = 17
    ROLL = 18


movements_values = [item.value for item in Movements]


def midi_setup_and_teardown(func):
    """Sets up midi before calling a function and tears it down after running

    Alternatively `pygame.midi.init()` and `pygame.midi.quit()` can be called at the beginning and end of each
    function manually
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        pygame.midi.init()
        results = func(*args, **kwargs)
        pygame.midi.quit()
        return results

    return wrapper


@midi_setup_and_teardown
def print_device_info():
    """Prints all midi devices that were found"""
    for i in range(pygame.midi.get_count()):
        interf, name, is_input, output, opened = pygame.midi.get_device_info(i)

        in_out = ""
        if is_input:
            in_out = "(input)"
        if output:
            in_out = "(output)"

        print(f"{i}: interface {interf}, name {name}, opened {opened} {in_out}")


@midi_setup_and_teardown
def wave_input_device_info() -> Optional[Tuple[int, str]]:
    """Automatically find input `Wave` input device"""
    result = None
    for i in range(pygame.midi.get_count()):
        interf, name, is_input, output, opened = pygame.midi.get_device_info(i)

        name_decoded = name.decode()
        if is_input and "Wave" in name_decoded:
            result = i, name_decoded
    return result


@midi_setup_and_teardown
def input_main(device_id):
    pygame.init()
    # Define our input device
    input_midi = pygame.midi.Input(device_id)
    pygame.display.set_mode((1, 1))
    going = True
    print_vals_results = ["", "", ""]

    while going:
        for event in pygame.event.get():
            if event.type in [pygame.QUIT]:
                going = False
            if event.type in [pygame.KEYDOWN]:
                # Program will stop if any button is pressed
                going = False
            if event.type in [pygame.midi.MIDIIN, pygame.USEREVENT]:
                print(event)

        # Check if there is any data in the buffer of the midi-device. If not, continue with the loop
        if not input_midi.poll():
            continue

        # Read all the data from the midi buffer (max_size=1024)
        midi_events = input_midi.read(1024)
        midi_events = pygame.midi.midis2events(midi_events, input_midi.device_id)

        # 3 different functions to use. Uncomment different ones to try out
        # post_custom_events(midi_events)
        # post_midi_events(midi_events, Movements.ROLL)
        print_vals_results = print_values(midi_events, print_vals_results)

    del input_midi


def print_values(midi_events, previous_vals: list) -> list:
    """Prints a 'data' midi event i.e. where there is information sent about roll/pitch/yaw

    If an event of a certain type is not found, prints the last value for that event
    """
    s = previous_vals
    for m_e in midi_events:
        if m_e.status == MidiStatus.DATA and m_e.data1 in movements_values:
            s = [m_e.data2 if m_e.data1 == val else prev_val for prev_val, val in zip(previous_vals, movements_values)]
    print(f"P: {s[0]:>3}, Y: {s[1]:>3}, R: {s[2]:>3}")
    return s


def post_custom_events(midi_events):
    """Only print certain events"""
    for m_e in midi_events:
        if m_e.status == MidiStatus.DATA and m_e.data1 == Movements.YAW:
            if m_e.data2 > 100:
                pygame.event.post(pygame.event.Event(pygame.USEREVENT, {"val": "down"}))
            if m_e.data2 < 30:
                pygame.event.post(pygame.event.Event(pygame.USEREVENT, {"val": "up"}))


def post_midi_events(midi_events, movement_type):
    """Post all midi events corresponding to data coming in"""
    for m_e in midi_events:
        if m_e.status == MidiStatus.DATA and m_e.data1 == movement_type:
            pygame.event.post(m_e)


def main():
    print("Make sure your Wave is connected to your computer as a midi device")
    print_device_info()
    device_id, name = wave_input_device_info()
    if device_id is not None:
        print(f"Found input device: {name} - {device_id}")
        print("Press any keyboard button to stop")
        input_main(device_id)
    else:
        print("Did not find a Wave device")


if __name__ == "__main__":
    main()



from queue import Queue
from typing import Optional, Any


class QueueWithPop(Queue):
    """A Queue that implements convenience methods"""

    def __init__(self, maxsize=0):
        super().__init__(maxsize=maxsize)

    def pop(self) -> Optional[Any]:
        """'safe' `pop`. Returns `None` if the queue is empty"""
        if self.qsize() != 0:
            val = self.get()
        else:
            val = None

        return val

    def pop_all(self):
        """Fetches everything on the Queue"""
        results = []
        while self.qsize() > 0:
            results.append(self.pop())
        return results



import asyncio
import threading
from typing import Callable, Optional

import serial
from serial import Serial
from serial.threaded import ReaderThread

from genki_wave.constants import BAUDRATE
from genki_wave.data.writing import get_start_api_package
from genki_wave.protocols import ProtocolThread, bluetooth_task, CommunicateCancel


from genki_wave.utils import get_serial_port, get_or_create_event_loop
from genki_wave.asyncio_runner import producer_bluetooth


class ReaderThreadSerial(ReaderThread):
    """A thin wrapper around `serial.threaded.ReaderThread`"""

    def __init__(self, serial_instance: Serial, protocol_factory: Callable):
        super().__init__(serial_instance, protocol_factory)
        self.write(get_start_api_package())

    @classmethod
    def from_port(cls, serial_port: Optional[str] = None) -> "ReaderThreadSerial":
        """Create a `ReaderThreadSerial` object from a serial port

        Args:
            serial_port: The serial port to read from. If `None` will automatically determine the port based on the
                operating system.

        Returns:
            An instance of `ReaderThreadSerial` with the specified port
        """
        port = get_serial_port() if serial_port is None else serial_port
        serial_instance = Serial(port, BAUDRATE, parity=serial.PARITY_EVEN)
        return cls(serial_instance, ProtocolThread)


class ReaderThreadBluetooth(threading.Thread):
    """An imitation of `serial.threaded.ReaderThread` for the bluetooth interface

    Note: Methods that clean up the connection are not fully worked out

    Note: Since the library used to interface with bluetooth only has an interface using `asyncio`, this
          uses the async code in a separate thread and then gets the data from there.
    """

    def __init__(self, ble_address, protocol_factory):
        super().__init__(daemon=True)
        self.protocol_factory = protocol_factory
        self.protocol = None
        self.alive = True
        self._lock = threading.Lock()
        self._comm = CommunicateCancel()
        self._ble_address = ble_address

    @classmethod
    def from_address(cls, ble_address: str) -> "ReaderThreadBluetooth":
        return cls(ble_address, ProtocolThread)

    def stop(self):
        """Stop the reader thread"""
        self.alive = False
        self.join(2)

    @staticmethod
    def run_producer(protocol, comm, ble_address) -> None:
        """Runs a producer fetching the data

        Args:
            comm: An object that allows `producer` and `consumer` to communicate when to cancel a process
            protocol: An object that knows how to process the raw data sent from the Wave ring into a structured format
                  and passes it along
            ble_address: Address of the bluetooth device to connect to. E.g. 'D5:73:DB:85:B4:A1'. If `None` it
                         connects via serial
        """
        loop = get_or_create_event_loop()
        producer = producer_bluetooth(protocol, comm, ble_address)
        tasks = asyncio.gather(*[producer])
        loop.run_until_complete(tasks)

    def run(self):
        """Reader loop"""
        self.protocol = self.protocol_factory()
        self.run_producer(self.protocol, self._comm, self._ble_address)

    def close(self):
        with self._lock:
            self._comm.cancel = True
            self.stop()

    def __enter__(self):
        """Enter context: Start code"""
        self.start()
        return self.protocol

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Leave context: close port"""
        self.close()


class WaveListener(threading.Thread):
    def __init__(self, ble_address, callbacks):
        self.ble_address = ble_address
        self.callbacks = callbacks
        super().__init__()

    def run(self):
        self.comm = CommunicateCancel()
        task = bluetooth_task(self.ble_address, self.comm, self.callbacks)
        loop = get_or_create_event_loop()
        loop.run_until_complete(task)

    def stop(self):
        self.comm.cancel = True

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.stop()



import abc
import csv
from pathlib import Path
from typing import Union, Optional, TextIO

from genki_wave.data.organization import ButtonEvent, DataPackage


class WaveCallback(abc.ABC):
    @abc.abstractmethod
    def _button_handler(self, data: ButtonEvent) -> None:
        pass

    @abc.abstractmethod
    def _data_handler(self, data: DataPackage) -> None:
        pass

    def __call__(self, data: Union[ButtonEvent, DataPackage]) -> None:
        if isinstance(data, ButtonEvent):
            self._button_handler(data)
        elif isinstance(data, DataPackage):
            self._data_handler(data)
        else:
            raise ValueError(f"Got data of unexpected type {type(data)}")


class ButtonAndDataPrint(WaveCallback):
    """
    Callback that prints out all button presses received and prints a data package every `print_data_every_n_seconds`
    seconds. Useful for debugging and testing.

    Args:
        print_data_every_n_seconds: The interval to print a data package
    """

    def __init__(self, print_data_every_n_seconds: Optional[float] = None):
        self._last_time = None
        self._print_data_every_n_seconds = print_data_every_n_seconds

    def _button_handler(self, data: ButtonEvent) -> None:
        # We use `str` to force the `enum` to print the long version of the name e.g. `ButtonId.MIDDLE` instead of `1`
        print(f"Button: {str(data.button_id)}, Action: {str(data.action)}")

    def _data_handler(self, data: DataPackage) -> None:
        """If there are more than `_print_data_every_n_seconds` seconds since something was printed out, print it"""
        if self._print_data_every_n_seconds is None:
            return

        if self._last_time is None:
            self._last_time = data.timestamp_us

        if (data.timestamp_us - self._last_time) > self._print_data_every_n_seconds * 10 ** 6:  # s to us
            print(data)
            self._last_time = data.timestamp_us


class CsvOutput(WaveCallback):
    """
    Exports the streaming data to a csv file, flushing it to the file every `flush_len` elements

    Default behaviour is to append the data if a file already exists

    Args:
        filename: The file the data will be exported to
        flush_len: How many samples to buffer before writing to a file
    """

    def __init__(self, filename: Path, flush_len: int = 256):

        self._events = []
        self._filename = filename
        self._fieldnames = DataPackage.flat_keys()
        self._flush_len = flush_len

        if not self._filename.exists():
            with open(self._filename, "w") as f:
                writer = self._get_writer(f)
                writer.writeheader()

    def _get_writer(self, f: TextIO) -> csv.DictWriter:
        return csv.DictWriter(f, fieldnames=self._fieldnames)

    def _reset_state(self) -> None:
        self._events = []

    def _button_handler(self, data: ButtonEvent) -> None:
        pass

    def _data_handler(self, data: DataPackage) -> None:
        """Receives the data and writes it out if enough data points have been collected"""
        self._events.append(data)

        if len(self._events) < self._flush_len:
            return

        with open(self._filename, "a") as f:
            writer = self._get_writer(f)

            for event in self._events:
                d = event.as_flat_dict()
                assert set(d.keys()) == set(writer.fieldnames), "Expected event.fieldnames == writer.fieldnames"
                writer.writerow(d)

        self._reset_state()



import argparse
import time
from typing import Union

from genki_wave.data.organization import ButtonEvent
from genki_wave.discover import run_discover_bluetooth
from genki_wave.threading_runner import ReaderThreadBluetooth, ReaderThreadSerial


class Sleeper:
    """A simple class to sleep for an amount of time that makes sure each loop runs for `sleep_for_x_seconds` seconds"""

    def __init__(self, sleep_for_x_seconds: float):
        self._last_time = None
        self._sleep_for_x_seconds = sleep_for_x_seconds

    def sleep(self):
        curr_time = time.time()
        sleep_time = max(0.0, self._sleep_for_x_seconds - (self.last_time - curr_time))
        time.sleep(sleep_time)
        self._last_time = time.time()

    @property
    def last_time(self):
        if self._last_time is None:
            self._last_time = time.time()
        return self._last_time


def main(reader_thread: Union[ReaderThreadBluetooth, ReaderThreadSerial], fetch_data_every_x_seconds: float):
    """A simple function to showcase reading from a thread.

    Fetches data every `fetch_data_every_x_seconds` and prints out all button presses in that timespan and the
    last value (usually a data package)
    """

    s = Sleeper(fetch_data_every_x_seconds)
    with reader_thread as wave:
        while True:
            val = wave.queue.pop_all()
            if val:
                for v in val:
                    if isinstance(v, ButtonEvent):
                        print(v)
                print(val[-1])

            s.sleep()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--ble-address", type=str)
    parser.add_argument("--use-serial", action="store_true")
    parser.add_argument("--fetch-data-every-x-seconds", type=float, default=1.0)
    args = parser.parse_args()

    if args.use_serial:
        main(ReaderThreadSerial.from_port(), args.fetch_data_every_x_seconds)
    else:
        if args.ble_address is None:
            print("No bluetooth address (--ble-address) supplied, searching for devices...")
            run_discover_bluetooth()
        else:
            print("Turn off using Ctrl + C (`KeyboardInterrupt`)")
            main(ReaderThreadBluetooth.from_address(args.ble_address), args.fetch_data_every_x_seconds)



import struct

from cobs import cobs

from genki_wave.data.organization import DeviceMode, PackageId, PackageMetadata, PackageType


def pad_with_zero_byte(b: bytes) -> bytes:
    return b + struct.pack("<B", 0)


def create_package_to_write(p: PackageMetadata, data: bytes) -> bytes:
    """Package the data in the correct format before sending it to a Wave ring"""
    # TODO(robert): `data` should probably be a class that has a `to_bytes` method
    if p.payload_size != len(data):
        raise ValueError(
            f"Expected payload_size and the size of the data to match. Got {p.payload_size} and {len(data)}"
        )
    b_tot = p.to_bytes() + data
    b_encoded = cobs.encode(b_tot)
    b_buffer = pad_with_zero_byte(b_encoded)
    return b_buffer


def get_start_api_package():
    """Get the package that puts the Wave ring into 'API mode' when it's sent to the device"""
    return create_package_to_write(
        PackageMetadata(type=PackageType.REQUEST, id=PackageId.DEVICE_MODE, payload_size=1),
        struct.pack("<B", DeviceMode.API.value),
    )



import abc
import asyncio
import logging
import struct
from queue import Queue
from typing import Optional, Union, List, Callable

from cobs import cobs
from serial.threaded import Packetizer
from bleak import BleakClient

from genki_wave.data.data_structures import QueueWithPop
from genki_wave.constants import API_CHAR_UUID
from genki_wave.data.writing import get_start_api_package
from genki_wave.data.organization import ButtonEvent, ButtonId, ButtonAction, DataPackage, process_byte_data
from genki_wave.utils import get_or_create_event_loop


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class CommunicateCancel:
    """
    Class that handles how to cancel asyncio loops with a button press and how to communicate it. Usually defined as
    a global and sends messages to each part of the script
    """

    cancel = False

    @staticmethod
    def is_cancel(button_event: Union[ButtonEvent, DataPackage]) -> bool:
        """Checks for a hard coded cancel event"""
        if isinstance(button_event, DataPackage):
            return False
        return button_event.button_id == ButtonId.TOP and button_event.action == ButtonAction.EXTRALONG


class ProtocolAbc(abc.ABC):
    """A protocol decodes raw data and connects producers and consumers af data from the input device

    It's a common abstraction that might be slightly abused here.

    A protocol:
        1) It defines methods that take the raw data-stream and transform it into a useful representation
        2) Since the methods that process the data-stream usually have to register the transformed data in
           a queue, these protocols defined here also have an instance of a Queue and allow consumer and
           producer, or different threads, to communicate.

    The flow of data is `data_received` -> `handle_packet`, see docstrings for explanations
    """

    @abc.abstractmethod
    def data_received(self, data: Union[bytearray, bytes]) -> None:
        """Gets the raw stream into a buffer and splits it up if needed"""
        pass

    @abc.abstractmethod
    def handle_packet(self, packet: Union[bytearray, bytes]) -> None:
        """Decodes and transforms the data into a useful representation and registers it in a queue"""
        pass

    @property
    @abc.abstractmethod
    def queue(self) -> Union[asyncio.Queue, Queue]:
        """The 'global' queue that is used to communicate between producer/consumer or between threads"""
        pass


def _handle_packet(packet: Union[bytearray, bytes]) -> Optional[Union[ButtonEvent, DataPackage]]:
    try:
        data = cobs.decode(packet)
        data = process_byte_data(data)
    except cobs.DecodeError:
        logger.debug("Got an exception decoding serial packet", exc_info=True)
        return None
    except (struct.error, ValueError):
        # Sometimes in the beginning we get garbage data
        logger.debug("Got input data error", exc_info=True)
        return None

    return data


class ProtocolAsyncio(ProtocolAbc, Packetizer):
    """Defines how to handle the bytes from the serial connection

    Note: This is a slight abuse of subclassing since we are pulling in more functionality than is needed
    """

    def __init__(self):
        super().__init__()
        get_or_create_event_loop()
        self._queue = asyncio.Queue()

    async def data_received(self, data: Union[bytearray, bytes]) -> None:
        """Buffer received data, find TERMINATOR, call handle_packet"""
        self.buffer.extend(data)
        while self.TERMINATOR in self.buffer:
            packet, self.buffer = self.buffer.split(self.TERMINATOR, 1)
            await self.handle_packet(packet)

    async def handle_packet(self, packet: Union[bytearray, bytes]) -> None:
        data = _handle_packet(packet)
        if data is None:
            return
        await self.queue.put(data)

    @property
    def queue(self) -> asyncio.Queue:
        return self._queue


class ProtocolThread(ProtocolAbc, Packetizer):
    def __init__(self):
        super().__init__()
        self._queue = QueueWithPop()

    def data_received(self, data: Union[bytearray, bytes]) -> None:
        """Buffer received data, find TERMINATOR, call handle_packet"""
        self.buffer.extend(data)
        while self.TERMINATOR in self.buffer:
            packet, self.buffer = self.buffer.split(self.TERMINATOR, 1)
            self.handle_packet(packet)

    def handle_packet(self, packet: Union[bytearray, bytes]) -> None:
        data = _handle_packet(packet)
        if data is None:
            return
        self.queue.put(data)

    @property
    def queue(self) -> Queue:
        return self._queue


def prepare_protocol_as_bleak_callback_asyncio(protocol: ProtocolAsyncio) -> Callable:
    async def _inner(sender: str, data: bytearray) -> None:
        # NOTE: `bleak` expects a function with this signature
        await protocol.data_received(data)

    return _inner


async def bluetooth_task(ble_address: str, comm: CommunicateCancel, callbacks: List[Callable]) -> None:
    protocol = ProtocolAsyncio()
    callback = prepare_protocol_as_bleak_callback_asyncio(protocol)
    print(f"Connecting to wave at address {ble_address}")
    async with BleakClient(ble_address) as client:
        await client.start_notify(API_CHAR_UUID, callback)
        await client.write_gatt_char(API_CHAR_UUID, get_start_api_package(), False)

        print("Connected to Wave")
        while True:
            package = await protocol.queue.get()

            if comm.is_cancel(package) or comm.cancel:
                print("Got a cancel message, exiting.")
                comm.cancel = True
                break

            for callback in callbacks:
                callback(package)

        await client.stop_notify(API_CHAR_UUID)



import asyncio
import glob
import sys

import serial


LINUX = "linux"
DARWIN = "darwin"


def get_system_name() -> str:
    if sys.platform.startswith(LINUX):
        return LINUX
    elif sys.platform.startswith(DARWIN):
        return DARWIN
    else:
        raise OSError(f"Unsupported system for serial reading: {sys.platform}")


SYSTEM_TO_SERIAL_PORT_GLOB = {LINUX: "/dev/ttyUSB*", DARWIN: "/dev/tty.usbserial*"}


def get_serial_port() -> str:
    """Get the serial port based on the system"""
    serial_port_glob = SYSTEM_TO_SERIAL_PORT_GLOB[get_system_name()]
    serial_ports = sorted(glob.glob(serial_port_glob))

    if not serial_ports:
        raise serial.SerialException("No serial port found. Is your serial cable plugged in?")

    return serial_ports[0]


DEFAULT_BLE_ADDRESS = {LINUX: "D5:73:DB:85:B4:A1", DARWIN: "b171e34e-9454-4d6d-b3d0-8740b703b66e"}[get_system_name()]


def get_or_create_event_loop():
    """Gets the running event loop or creates a new one and returns it

    By default `asyncio` only starts an event loop in the main thread, so when running in another thread we
    need to explicitly create a new event loop for that particular thread.
    """
    try:
        return asyncio.get_event_loop()
    except RuntimeError as ex:
        if "There is no current event loop in thread" in str(ex):
            asyncio.set_event_loop(asyncio.new_event_loop())
            return asyncio.get_event_loop()



from setuptools import setup, find_packages


with open("README.md") as f:
    readme = f.read()

requires = ["bleak", "cobs", "pyserial", "pyserial-asyncio"]

setup(
    name="genki-wave",
    version="0.1.6",
    description="Python API for Wave by Genki",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Robert Torfason",
    author_email="robert@genkiinstruments.com",
    url="https://github.com/genkiinstruments/genki-wave",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires=">=3.8",
    packages=find_packages(exclude=("tests", "docs")),
    install_requires=requires,
)
